from dune.perftool.pdelab.argument import pymbolic_coefficient
from dune.perftool.generation import (domain,
                                      get_counter,
                                      iname,
                                      instruction,
                                      silenced_warning,
                                      )
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.pdelab.spaces import name_lfs
from dune.perftool.sumfact.amatrix import (AMatrix,
                                               quadrature_points_per_direction,
                                               basis_functions_per_direction,
                                               name_theta,
                                               name_theta_transposed,
                                               )
from dune.perftool.loopy.stages import stage_insn
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )

from loopy import Reduction

from pytools import product


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


def setup_theta(element, container, restriction, component, a_matrices):
    number_basis = product(mat.cols for mat in a_matrices)
    shape = (number_basis,)
    inp = get_buffer_temporary("buffer",
                               shape=shape)
    silenced_warning('read_no_write({})'.format(inp))

    # Write initial coefficients into buffer
    basisiname = sumfact_iname(number_basis, "basis")
    lfs = name_lfs(element, restriction, component)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),))
    return instruction(assignee=assignee,
                       expression=coeff,
                       depends_on=frozenset({stage_insn(0)}),
                       )


# TODO this code is WIP and mainly used for experiments.
def start_sumfactorization(element, container, restriction, component):
    theta = name_theta()
    rows = quadrature_points_per_direction()
    cols = basis_functions_per_direction()
    a_matrix = AMatrix(theta, rows, cols)
    a_matrices = (a_matrix, a_matrix)

    # Do stage 1
    initialize_buffer("buffer",
                      base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                      num=2
                      )

    insn_dep = setup_theta(element, container, restriction, component, a_matrices)

    sum_factorization_kernel(a_matrices, "buffer", 0, frozenset({insn_dep}))

    # Do stage 3 (for f=u => mass matrix)
    theta_transposed = name_theta_transposed()
    a_matrix_transposed = AMatrix(theta_transposed, cols, rows)
    a_matrices_transposed = (a_matrix_transposed, a_matrix_transposed)

    return sum_factorization_kernel(a_matrices_transposed, "buffer", 2)


def sum_factorization_kernel(a_matrices, buffer, stage, insn_dep=frozenset({})):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buffer: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input.
    insn_dep: an instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    """
    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.cols, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        inp = get_buffer_temporary(buffer,
                                   shape=inp_shape,
                                   dim_tags="f,f")

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buffer)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.rows, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        out = get_buffer_temporary(buffer,
                                   shape=out_shape,
                                   dim_tags="c,c")

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")
        k = sumfact_iname(a_matrix.cols, "red")

        # Construct the matrix-matrix-multiplication expression a_ik*in_kj
        from dune.perftool.sumfact.amatrix import ColMajorAccess
        prod = Product((Call(ColMajorAccess(a_matrix.a_matrix), (Variable(i), Variable(k))),
                        Subscript(Variable(inp), (Variable(k), Variable(j)))
                        ))

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=Subscript(Variable(out), (Variable(i), Variable(j))),
                                          expression=Reduction("sum", k, prod),
                                          forced_iname_deps=frozenset({i, j}),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep.union(frozenset({stage_insn(stage)})),
                                          )
                              })

    return out
